"""
MIT License

Copyright (c) 2023 Konrad (@konradsic)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime
from typing import Any, Dict, List, Literal, Optional, Union

from .. import utils as utils
from ..enums import Languages
from ..errors import (AccessDenied, APIKeyDisabled, APILimitExceeded,
                      InternalApplicationError, InvalidAPIKey, InvalidDate,
                      NoLocationFound, WeatherAPIException)
from ..responses import (AstronomicalData, CurrentWeatherData, ForecastData,
                         FutureData, IPData, LocationData, SportsData)
from .core import BaseAPIClient

WEATHERAPI_BASE_URL = "https://api.weatherapi.com/v1/"
BOOL_REPLACE = {True: "yes", False: "no"}

__all__ = (
    "Client",
)

class Client(BaseAPIClient):
    """
    A WeatherAPI.com client for fetching various weather information

    Parameters
    ----------
    api_key: :class:`str`
        API Key used to authenticate when sending requests
    lang: Optional[ Union[:class:`str`, :class:`Languages`] ]
        Language from the :class:`Languages` enum or a string representing the language or language code (preferably).
        To get a list of languages visit :class:`Languages`. If ``None`` then the default language is used (English)
    dt: Optional[:class:`int`]
        Restrict date output for Forecast and History API method. (Required for History and Future API)
    end_dt: Optional[:class:`int`]
        Restrict date output for History API method. Only works for API on Pro plan and above. (Available for History API)
    hour: Optional[:class:`int`]
        Restricting forecast or history output to a specific hour in a given day.
    aqi: :class:`bool`
        Enable/Disable Air Quality data in forecast API output. Defaults to "no".
    tides: :class:`bool`
        Enable/Disable Tide data in Marine API output. Defaults to "no".
    kwargs: Dict[:class:`str`, Any]
        Additional keyword arguments passed by default to requests made by the client
        
    Attributes
    -------------
    lang: Optional[:class:`Languages`]
        Language from the :class:`Languages` enum or a string representing the language or language code (preferably).
        To get a list of languages visit :class:`Languages`. If ``None`` then the default language is used (English)
    dt: Optional[:class:`int`]
        Restricted date output for Forecast and History API method. (Required for History and Future API)
    end_dt: Optional[:class:`int`]
        Restricted date output for History API method. Only works for API on Pro plan and above.
    hour: Optional[:class:`int`]
        Restricted forecast or history output to a specific hour in a given day.
    aqi: :class:`bool`
        Indicates if Air Quality data has been enabled
    tides: :class:`bool`
        Indicates if tides data in the Marine API has been enabled
    kwargs: Dict[:class:`str`, Any]
        Additional keyword arguments passed by default to requests made by the client
    """
    def __init__(
        self,
        api_key: str,
        lang: Optional[Union[str, Languages]] = None,
        dt: Optional[int] = None,
        end_dt: Optional[int] = None,
        hour: Optional[int] = None,
        aqi: bool = False,
        tides: bool = False,
        **kwargs: Dict[str, Any]
    ) -> None:
        lang_code = None
        if lang is not None: 
            lang_code = utils.find_language(lang, asobj=True)
        opts = {
            "key": api_key,
            **kwargs
        }
        if lang_code: opts.update(lang = lang_code.value if lang_code is not None else None)

        super().__init__(base_url=WEATHERAPI_BASE_URL,
                         default_options=opts)
        
        self.dt = dt
        self.end_dt = end_dt
        self.hour = hour
        self.aqi = aqi
        self.tides = tides
        self.kwargs = kwargs
        self.lang = Languages(lang_code) if lang_code else None
    
    
    def _call_request(self, endpoint: str, options: Dict[str, Any]) -> Dict[str, Any]:
        """Private method used to make requests to WeatherAPI"""
        final_options = self.default_options.copy()

        for k,v in final_options.items(): # check - remove NoneType values
            if v is None: del final_options[k]
        # add options to final_options
        for k,v in options.items():
            # also - replace bool with "yes"/"no"
            if v is not None: final_options[k] = BOOL_REPLACE.get(v, v)

        resp = self._request(endpoint, **final_options)

        if not resp[1].status_code < 400:
            # raise errors yay
            error_data = resp[0]["error"]
            code = error_data["code"]
            status = resp[1].status_code
            msg = error_data["message"]

            if code == 1006: raise NoLocationFound(status, code, msg)
            elif code == 2006: raise InvalidAPIKey(status, code, msg)
            elif code == 2007: raise APILimitExceeded(status, code, msg)
            elif code == 2008: raise APIKeyDisabled(status, code, msg)
            elif code == 2009: raise AccessDenied(status, code, msg)
            elif code == 9999: raise InternalApplicationError(status, code, msg)
            else: raise WeatherAPIException(status, code, msg)

        return resp
    
    def event(self):
        """A decorator that turns a function into an event. For example

        .. code:: python

            import weatherly
            client = weatherly.Client(api_key=...)

            @client.event
            def on_error(payload):
                print(f"An error occured! Payload: {payload}")
        
        In the example above, by adding ``@client.event`` the ``on_error`` function has turned into an error handler function
        """
        raise NotImplementedError

    def set_language(self, lang: Union[str, Languages]) -> Optional[Languages]:
        """
        Set client's language when requesting data.
        
        Parameters
        -----------
        lang: Union[:class:`str`, :class:`Languages`]
            Language to set. Can be either a string that is lanuage's name or code or a :class:`Languages` enum object.
            
        Returns
        ---------
        Optional[:class:`Languages`]
            An enum class representing the language of the client. Is ``None`` when something went wrong and the language was not set.
        """
        lang_class = utils.find_language(lang, asobj=True)
        if not lang_class:
            return None

        self.lang = lang_class 
        return self.lang

    def get_current_weather(self, 
        query: str, 
        *,
        lang: Optional[Union[str, Languages]] = None,
        aqi: Optional[bool] = None,
        **kwargs: Dict[str, Any]
    ) -> CurrentWeatherData:
        """
        Get current weather data

        Parameters
        ----------
        query: :class:`str`
            Query string - location you want to get weather data for
        lang: Optional[Union[:class:`str`, Languages]]
            Language from the :class:`Languages` enum or a string representing the language or language code (preferably).
            To get a list of languages visit :class:`Languages`.
        aqi: Optional[:class:`bool`]
            Enable/Disable Air Quality data in forecast API output. If nothing is passed, then it defaults to client default value.
        kwargs: Dict[:class:`str`, Any]
            Additional keyword arguments to request

        Returns
        -------
        :class:`CurrentWeatherData`
            Current weather data class

        Raises
        ---------
        :exc:`NoLocationFound`
            Raised when no location for given query was found
        :exc:`InvalidAPIKey`
            Raised when the API key is invalid
        :exc:`APILimitExceeded`
            Raised when API key calls limit was exceeded
        :exc:`APIKeyDisabled`
            Raised when API key is disabled
        :exc:`AccessDenied`
            Raised when access to given resource was denied
        :exc:`InternalApplicationError`
            Raised when there was a very rare internal application error
        :exc:`WeatherAPIException`
            Raised when something else went wrong, that does not have a specific exception class.
        """
        options = {
            "aqi": aqi or self.aqi,
            "q": query,
            **kwargs
        }
        if lang is not None: options["lang"] = lang
        resp = self._call_request("current.json", options)

        weather = CurrentWeatherData(resp[0], resp[1].status_code, None)
        return weather

    def get_locations(self, query: str):
        """
        Get locations for given query

        Parameters
        ---------------
        query: :class:`str`
            Query string, a location you are searching for

        Returns
        -----------
        List[:class:`LocationData`]
            A list of :class:`LocationData` classes.

        Raises
        ---------
        :exc:`NoLocationFound`
            Raised when no location for given query was found
        :exc:`InvalidAPIKey`
            Raised when the API key is invalid
        :exc:`APILimitExceeded`
            Raised when API key calls limit was exceeded
        :exc:`APIKeyDisabled`
            Raised when API key is disabled
        :exc:`AccessDenied`
            Raised when access to given resource was denied
        :exc:`InternalApplicationError`
            Raised when there was a very rare internal application error
        :exc:`WeatherAPIException`
            Raised when something else went wrong, that does not have a specific exception class.
        """
        resp = self._call_request("search.json",{"q": query})

        locations = []
        for loc in resp[0]:
            locations.append(LocationData(loc, resp[1].status_code, None))
        return locations

    def get_forecast_data(
        self,
        query: str, 
        days: int,
        *,
        aqi: Optional[bool] = None,
        alerts: Optional[bool] = None,
        lang: Optional[Union[str, Languages]] = None,
        **kwargs: Dict[str, Any]
    ) -> ForecastData:
        """
        Get forecast data from Forecast API

        Parameters
        -------------
        query: :class:`str`
            Query string, location you want to get forecast data for
        days: :class:`int`
            Number of days of weather forecast. Value ranges from 1 to 10
        aqi: Optional[:class:`bool`]
            Enable/Disable Air Quality data. Defaults to ``None`` (will use client's default)
        alerts: Optional[:class:`bool`]
            Enable/Disable alerts data. Defaults to ``None`` (will use client's default)
        lang: Optional[Union[:class`str`, :class`Languages`]]
            Language from the :class:`Languages` enum or a string representing the language or language code (preferably).
            To get a list of languages visit :class:`Languages`.
        kwargs: Dict[:class:`str`, Any]
            Additional keyword arguments that will be passed to the request.
            
        Returns
        ----------
        :class:`ForecastData`
            Fetched forecast data as a class.

        Raises
        ---------
        :exc:`NoLocationFound`
            Raised when no location for given query was found
        :exc:`InvalidAPIKey`
            Raised when the API key is invalid
        :exc:`APILimitExceeded`
            Raised when API key calls limit was exceeded
        :exc:`APIKeyDisabled`
            Raised when API key is disabled
        :exc:`AccessDenied`
            Raised when access to given resource was denied
        :exc:`InternalApplicationError`
            Raised when there was a very rare internal application error
        :exc:`WeatherAPIException`
            Raised when something else went wrong, that does not have a specific exception class.
        """
        options = {
            "aqi": aqi or self.aqi,
            "q": query,
            "alerts": alerts or self.kwargs.get("alerts"),
            "days": days,
            **kwargs
        }
        if lang is not None: options["lang"] = lang

        resp = self._call_request("forecast.json", options)
        forecast = ForecastData(resp[0], resp[1].status_code, None)
        return forecast

    def get_historical_data(
        self,
        query: str,
        date: str,
        *,
        aqi: Optional[bool] = None,
        alerts: Optional[bool] = None,
        lang: Optional[Union[str, Languages]] = None,
        **kwargs: Dict[str, Any]
    ) -> ForecastData:
        """
        Retrieve historical data for given day and query. Uses History API.

        Parameters
        -----------------
        query: :class:`str`
            Query string, location you want to get forecast data for
        date: :class:`str`
            A date string in format yyyy-mm-dd representing the day you want to get data for
        aqi: Optional[:class:`bool`]
            Enable/Disable Air Quality data. Defaults to ``None`` (will use client's default)
        alerts: Optional[:class:`bool`]
            Enable/Disable alerts data. Defaults to ``None`` (will use client's default)
        lang: Optional[Union[:class`str`, :class`Languages`]]
            Language from the :class:`Languages` enum or a string representing the language or language code (preferably).
            To get a list of languages visit :class:`Languages`.
        kwargs: Dict[:class:`str`, Any]
            Additional keyword arguments that will be passed to the request.
        
        Returns
        ----------
        :class:`ForecastData`
            Forecast data for given day and query.

        Raises
        ---------
        :exc:`NoLocationFound`
            Raised when no location for given query was found
        :exc:`InvalidAPIKey`
            Raised when the API key is invalid
        :exc:`APILimitExceeded`
            Raised when API key calls limit was exceeded
        :exc:`APIKeyDisabled`
            Raised when API key is disabled
        :exc:`AccessDenied`
            Raised when access to given resource was denied
        :exc:`InternalApplicationError`
            Raised when there was a very rare internal application error
        :exc:`WeatherAPIException`
            Raised when something else went wrong, that does not have a specific exception class.
        :exc:`InvalidDate`
            Raised when the ``date`` parameter is invalid (doesn't match the format or isn't a date before (or) today)
        """
        options = {
            "aqi": aqi or self.aqi,
            "q": query,
            "alerts": alerts or self.kwargs.get("alerts"),
            "dt": date,
            **kwargs
        }
        if lang is not None: options["lang"] = lang

        # check if given date is really "historical"
        try:
            splitted = date.split("-")
            datetuple = datetime.datetime(
                int(splitted[0]), 
                int(splitted[1][1:]) if splitted[1].startswith("0") else int(splitted[1]), 
                int(splitted[2][1:]) if splitted[2].startswith("0") else int(splitted[2]),
                0,0)
            epoch = datetuple.timestamp()
        except Exception as exc:
            raise InvalidDate(f"Failed to convert date {date}: Invalid format") from exc

        now = datetime.datetime.timestamp(datetime.datetime.utcnow())

        if epoch > now: raise InvalidDate("Date should be before current time, switch from History API to Future to use future dates.")

        resp = self._call_request("history.json", options)
        history = ForecastData(resp[0], resp[1].status_code, None)
        return history
        
    def get_future_data(
        self,
        query: str,
        date: str,
        *,
        lang: Optional[Union[str, Languages]] = None,
        **kwargs: Dict[str, Any]
    ) -> FutureData:
        """
        Retrieve future data for given day and query. Uses Future API.

        Parameters
        -----------------
        query: :class:`str`
            Query string, location you want to get forecast data for
        date: :class:`str`
            A date string in format yyyy-mm-dd representing the day you want to get data for
        lang: Optional[Union[:class`str`, :class`Languages`]]
            Language from the :class:`Languages` enum or a string representing the language or language code (preferably).
            To get a list of languages visit :class:`Languages`.
        kwargs: Dict[:class:`str`, Any]
            Additional keyword arguments that will be passed to the request.
        
        Returns
        ----------
        :class:`FutureData`
            Future data for given day and query.

        Raises
        ---------
        :exc:`NoLocationFound`
            Raised when no location for given query was found
        :exc:`InvalidAPIKey`
            Raised when the API key is invalid
        :exc:`APILimitExceeded`
            Raised when API key calls limit was exceeded
        :exc:`APIKeyDisabled`
            Raised when API key is disabled
        :exc:`AccessDenied`
            Raised when access to given resource was denied
        :exc:`InternalApplicationError`
            Raised when there was a very rare internal application error
        :exc:`WeatherAPIException`
            Raised when something else went wrong, that does not have a specific exception class.
        :exc:`InvalidDate`
            Raised when the ``date`` parameter is invalid (doesn't match the format or isn't a date after today)
        """
        options = {
            "q": query,
            "dt": date,
            **kwargs
        }
        if lang is not None: options["lang"] = lang

        # check if given date is really "historical"
        try:
            splitted = date.split("-")
            datetuple = datetime.datetime(
                int(splitted[0]), 
                int(splitted[1][1:]) if splitted[1].startswith("0") else int(splitted[1]), 
                int(splitted[2][1:]) if splitted[2].startswith("0") else int(splitted[2]),
                0,0)
            epoch = datetuple.timestamp()
        except Exception as exc:
            raise InvalidDate(f"Failed to convert date {date}: Invalid format") from exc

        now = datetime.datetime.timestamp(datetime.datetime.utcnow())

        if epoch < now: raise InvalidDate("Date should be after current time, switch from Future API to History to use past dates.")

        resp = self._call_request("future.json", options)
        future = FutureData(resp[0], resp[1].status_code, None)
        return future
        
    def get_astronomical_data(
        self,
        query: str,
        date: str,
        **kwargs: Dict[str, Any]
    ) -> AstronomicalData:
        """
        Get astronomical data from Astronomy API

        Parameters
        -------------
        query: :class:`str`
            Query string, location you want to get forecast data for
        date: :class:`str`
            Date in format yyyy-MM-dd and on or after 1st Jan, 2010 (2010-01-01)
        kwargs: Dict[:class:`str`, Any]
            Additional keyword arguments that will be passed to the request.
            
        Returns
        ----------
        :class:`AstronomicalData`
            Fetched astronomical data as a class.

        Raises
        ---------
        :exc:`NoLocationFound`
            Raised when no location for given query was found
        :exc:`InvalidAPIKey`
            Raised when the API key is invalid
        :exc:`APILimitExceeded`
            Raised when API key calls limit was exceeded
        :exc:`APIKeyDisabled`
            Raised when API key is disabled
        :exc:`AccessDenied`
            Raised when access to given resource was denied
        :exc:`InternalApplicationError`
            Raised when there was a very rare internal application error
        :exc:`WeatherAPIException`
            Raised when something else went wrong, that does not have a specific exception class.
        """
        options = {
            "q": query,
            "dt": date,
            **kwargs
        }
        resp = self._call_request("astronomy.json", options)
        astro = AstronomicalData(resp[0], resp[1].status_code, None)
        return astro

    def get_marine_data(
        self,
        query: str,
        *,
        tides: Optional[bool],
        **kwargs: Dict[str, Any]
    ):
        raise NotImplementedError

    def get_ip_data(
        self,
        ip_address: str,
        **kwargs: Dict[str, Any]
    ) -> IPData:
        """
        Look for data for the given IP address.

        Parameters
        -------------
        ip_address: :class:`str`
            IP address you want to get data for. Can be ipv4 or ipv6
        kwargs: Dict[:class:`str`, Any]
            Additional keyword arguments that will be passed to the request.
            
        Returns
        ----------
        :class:`IPData`
            Fetched IP data as a class.

        Raises
        ---------
        :exc:`NoLocationFound`
            Raised when either IP address was invalid or no matching location was found.
        :exc:`InvalidAPIKey`
            Raised when the API key is invalid
        :exc:`APILimitExceeded`
            Raised when API key calls limit was exceeded
        :exc:`APIKeyDisabled`
            Raised when API key is disabled
        :exc:`AccessDenied`
            Raised when access to given resource was denied
        :exc:`InternalApplicationError`
            Raised when there was a very rare internal application error
        :exc:`WeatherAPIException`
            Raised when something else went wrong, that does not have a specific exception class.
        """
        options = {
            "q": ip_address,
            **kwargs
        }
        resp = self._call_request("ip.json", options)
        ip = IPData(resp[0], resp[1].status_code, None)
        return ip

    def get_sports_data(
        self,
        query: str,
        **kwargs: Dict[str, Any]
    ) -> SportsData:
        """
        Get sports data for a given query. Uses Sports API.

        Parameters
        -------------
        query: :class:`str`
            Query string, location you want to get sports data for
        kwargs: Dict[:class:`str`, Any]
            Additional keyword arguments that will be passed to the request.
            
        Returns
        ----------
        :class:`SportsData`
            Retrieved sports data as a class.

        Raises
        ---------
        :exc:`NoLocationFound`
            Raised when either IP address was invalid or no matching location was found.
        :exc:`InvalidAPIKey`
            Raised when the API key is invalid
        :exc:`APILimitExceeded`
            Raised when API key calls limit was exceeded
        :exc:`APIKeyDisabled`
            Raised when API key is disabled
        :exc:`AccessDenied`
            Raised when access to given resource was denied
        :exc:`InternalApplicationError`
            Raised when there was a very rare internal application error
        :exc:`WeatherAPIException`
            Raised when something else went wrong, that does not have a specific exception class.
        """
        options = {
            "q": query,
            **kwargs
        }
        resp = self._call_request("sports.json", options)
        sports = SportsData(resp[0], resp[1].status_code, None)
        return sports


    
    def __str__(self):
        return f"<{self.__class__.__name__} api_key={self.default_options['key']} lang={self.lang}>"
    
    def __repr__(self):
        return repr(self.__str__())
    


"""
MIT License

Copyright (c) 2023 Konrad (@konradsic)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from typing import (
    Dict,
    Any,
    Union,
    Optional,
    List, 
    Iterator,
    Literal,
)
from .enums import SportsEventType

__all__ = (
    "CurrentWeatherData",
    "LocationData",
    "AirQualityData",
    "ForecastData",
    "ForecastDay",
    "ForecastHour",
    "AlertData",
    "FutureData",
    "AstronomicalData",
    "IPData",
    "SportsEvent",
    "SportsData",
)


GB_DEFRA_BAND = ("Low", "Low", "Low", "Moderate", "Moderate", "Moderate", "High", "High", "High", "Very High")

class APIResponse:
    """
    Represents a basic response from Weather API
    
    Attributes
    ----------
    raw: Dict[:class:`str`, Any]
        Raw response in a JSON-like format (converted to a python dictionary)
    status: :class:`int`
        HTTP status of the response. 200 is OK, and is the most common status
    code: Optional[:class:`int`]
        Response code. In some cases this can be ``None``
    """
    def __init__(
        self,
        raw: Dict[str, Any],
        status: int,
        code: Optional[int]
    ) -> None:
        self.raw: Dict[str, Any] = raw
        self.status: int = status
        self.code: Optional[int] = code

class CurrentWeatherData(APIResponse):
    """
    Current weather data, a common return type from methods that requests this from WeatherAPI.com.
    
    Attributes
    ----------
    raw: Dict[:class:`str`, Any]
        Raw response in a JSON-like format (converted to a python dictionary)
    status: :class:`int`
        HTTP status of the response. 200 is OK, and is the most common status.
    code: Optional[:class:`int`]
        Response code. In some cases this can be ``None``
    location: :class:`LocationData`
        Location of the requested weather data
    last_updated_epoch: :class:`int`
        The timestamp when the weather data was last updated.
    temp_c: :class:`float`
        Weather temperature in Celsius
    temp_f: :class:`float`
        Weather temperature in Fahrenheit
    is_day: :class:`bool`
        Wherever the current time is considered day time.
    condition_text: :class:`str`
        Weather condition text
    condition_icon: :class:`str`
        Link to the weather condition icon
    condition_code: :class:`int`
        Code of current weather condition
    wind_mph: :class:`float`
        Current wind speed in miles per hour (mph)
    wind_kph: :class:`float`
        Current wind speed in kilometers per hour (kph)
    wind_degree: :class:`int`
        Direction of wind in degrees
    wind_dir: :class:`str`
        Direction of wind as a string (e.g. "W")
    pressure_mb: :class:`float`
        Pressure in millibars (mb)
    pressure_in: :class:`float`
        Pressure in inches (in)
    precip_mm: :class:`float`
        Precipation (water that is falling out of the sky) in millimeters (mm)
    precip_in: :class:`float`
        Precipation (water that is falling in the sky) in inches (in)
    humidity: :class:`int`
        Humidity in integer percentage
    cloud: :class:`int`
        Cloud cover as percentage
    feelslike_c: :class:`float`
        Feels like temperature in Celsius
    feelslike_f: :class:`float`
        Feels like temperature in Fahrenheit
    uv: :class:`float`
        UV Index
    aqi: Optional[:class:`AirQualityData`]
        Air quality data. Can be ``None`` (if this field of data was not requested)
    """
    def __init__(
        self,
        raw: Dict[str, Any],
        status: int,
        code: Optional[int]
    ) -> None:
        super().__init__(raw, status, code)
        
        self.location: LocationData = LocationData(raw["location"], status, code)
        raw = raw["current"]
        
        self.aqi: Optional[AirQualityData] = AirQualityData(raw["air_quality"], status, code) if raw.get("air_quality") else None
        self.last_updated_epoch: int = raw["last_updated_epoch"]
        self.temp_c: float = raw["temp_c"]
        self.temp_f: float = raw["temp_f"]
        self.is_day: bool = bool(raw["is_day"])
        self.condition_text: str = raw["condition"]["text"]
        self.condition_icon: str = raw["condition"]["icon"]
        self.condition_code: int = raw["condition"]["code"]
        self.wind_mph: float = raw["wind_mph"]
        self.wind_kph: float = raw["wind_kph"]
        self.wind_degree: int = raw["wind_degree"]
        self.wind_dir: str = raw["wind_dir"]
        self.pressure_mb: float = raw["pressure_mb"]
        self.pressure_in: float = raw["pressure_in"]
        self.precip_mm: float = raw["precip_mm"]
        self.precip_in: float = raw["precip_in"]
        self.humidity: int = raw["humidity"]
        self.cloud: int = raw["cloud"]
        self.feelslike_c: float = raw["feelslike_c"]
        self.feelslike_f: float = raw["feelslike_f"]
        self.uv: float = raw["uv"]

class LocationData(APIResponse):
    """
    Location data, returned with most requests.
    
    Attributes
    --------------
    raw: Dict[:class:`str`, Any]
        Raw response in a JSON-like format (converted to a python dictionary)
    status: :class:`int`
        HTTP status of the response. 200 is OK, and is the most common status.
    code: Optional[:class:`int`]
        Response code. In some cases this can be ``None``
    id: Optional[:class:`int`]
        A specific ID of the location. Can be ``None``
    name: :class:`str`
        Name of the location (e.g. London)
    region: :class:`str`
        A region of the location
    country: :class:`str`
        Country where the location is
    latitude: :class:`float`
        Latitude coordinate of the location
    longitude: :class:`float`
        Longitude coordinate of the location
    timezone_id: Optional[:class:`str`]
        Timezone ID of the location (e.g. Europe/London). Could be ``None`` when using the Search/Autocomplete API.
    localtime_epoch: Optional[:class:`int`]
        Local time of the location as a timestamp
    localtime_formatted: Optional[:class:`str`]
        Formatted local time of the location
    """
    def __init__(
        self, 
        raw: Dict[str, Any],
        status: int,
        code: Optional[int]
    ) -> None:
        super().__init__(raw, status, code)

        self.id: int = raw.get('id', None)
        self.name: str = raw['name']
        self.region: str = raw['region']
        self.country: str = raw['country']
        self.latitude: float = raw['lat']
        self.longitude: float = raw['lon']
        self.timezone_id: str = raw.get('tz_id', None)
        self.localtime_epoch: int = raw.get('localtime_epoch')
        self.localtime_formatted: str = raw.get('localtime', None)


class AirQualityData(APIResponse):
    """
    Attributes
    --------------
    raw: Dict[:class:`str`, Any]
        Raw response in a JSON-like format (converted to a python dictionary)
    status: :class:`int`
        HTTP status of the response. 200 is OK, and is the most common status.
    code: Optional[:class:`int`]
        Response code. In some cases this can be ``None``
    co: :class:`float`
        Carbon Monoxide (μg/m3)
    o3: :class:`float`
        Ozone (μg/m3)
    no2: :class:`float`	
        Nitrogen dioxide (μg/m3)
    so2: :class:`float`
        Sulphur dioxide (μg/m3)
    pm2_5: :class:`float`
        PM2.5 (μg/m3)
    pm10: :class:`float`
        PM10 (μg/m3)
    us_epa_index: :class:`int`
        US - EPA standard.
            * 1 means Good
            * 2 means Moderate
            * 3 means Unhealthy for sensitive group
            * 4 means Unhealthy
            * 5 means Very Unhealthy
            * 6 means Hazardous
    gb_defra_index: :class:`int`
        UK Defra Index
        
        +--------+------+-------+-------+----------+----------+----------+-------+-------+-------+------------+
        | Index  | 1    | 2     | 3     | 4        | 5        | 6        | 7     | 8     | 9     | 10         |
        +========+======+=======+=======+==========+==========+==========+=======+=======+=======+============+
        | Band   | Low  | Low   | Low   | Moderate | Moderate | Moderate | High  | High  | High  | Very High  |
        +--------+------+-------+-------+----------+----------+----------+-------+-------+-------+------------+
        | µgm^-3 | 0-11 | 12-23 | 24-35 | 36-41    | 42-47    | 48-53    | 54-58 | 59-64 | 65-70 | 71 or more |
        +--------+------+-------+-------+----------+----------+----------+-------+-------+-------+------------+
    
    gb_defra_band: :class:`str`
        A band corresponding to the :py:attr:`~gb_defra_index`
    """
    def __init__(
        self,
        raw: Dict[str, Any],
        status: int,
        code: Optional[int]
    ) -> None:
        super().__init__(raw, status, code)
        
        self.co: float = raw["co"]
        self.o3: float = raw["o3"]
        self.no2: float = raw["no2"]
        self.so2: float = raw["so2"]
        self.pm2_5: float = raw["pm2_5"]
        self.pm10: float = raw["pm10"]
        self.us_epa_index: int = raw["us-epa-index"]
        self.gb_defra_index: int = raw["gb-defra-index"]

        self.gb_defra_band: str = GB_DEFRA_BAND[self.gb_defra_index-1]

class AlertData(APIResponse):
    """
    An alert from WeatherAPI

    Attributes
    -------------
    raw: Dict[:class:`str`, Any]
        Raw response in a JSON-like format (converted to a python dictionary)
    status: :class:`int`
        HTTP status of the response. 200 is OK, and is the most common status.
    code: Optional[:class:`int`]
        Response code. In some cases this can be ``None``
    headline: :class:`str`
        Alert headline
    msg_type: :class:`str`
        Type of alert
    severity: :class:`str`
        Severity of alert
    urgency: :class:`str`
        Urgency
    areas: :class:`str`
        Areas covered
    category: :class:`str`
        Category
    certainty: :class:`str`
        Certainty
    event: :class:`str`
        Event
    note: :class:`str`
        Note
    effective: :class:`str`
        Effective
    expires: :class:`str`
        Expires
    description: :class:`str`
        Description
    instruction: :class:`str`
        Instruction
    """
    def __init__(
        self,
        raw: Dict[str, Any],
        status: int,
        code: Optional[int]
    ) -> None:
        super().__init__(raw, status, code)
        
        self.headline: str = raw["headline"]
        self.msg_type: str = raw["msgType"]
        self.severity: str = raw["severity"]
        self.urgency: str = raw["urgency"]
        self.areas: str = raw["areas"]
        self.category: str = raw["category"]
        self.certainty: str = raw["certainty"]
        self.event: str = raw["event"]
        self.note: str = raw["note"]
        self.effective: str = raw["effective"]
        self.expires: str = raw["expires"]
        self.description: str = raw["desc"]
        self.instruction: str = raw["instruction"]

class ForecastHour(APIResponse):
    """
    Forecast hour, an element of :class:`ForecastDay`
    
    Attributes
    -------------
    raw: Dict[:class:`str`, Any]
        Raw response in a JSON-like format (converted to a python dictionary)
    status: :class:`int`
        HTTP status of the response. 200 is OK, and is the most common status.
    code: Optional[:class:`int`]
        Response code. In some cases this can be ``None``
    time_epoch: :class:`int`
        Time as epoch
    time: :class:`str`
        Date and time
    temp_c: :class:`float`
        Temperature in celsius
    temp_f: :class:`float`
        Temperature in fahrenheit
    condition_text: :class:`str`
        Weather condition text
    condition_icon: :class:`str`
        Weather condition icons
    condition_code: :class:`int`
        Weather condition code
    wind_mph: :class:`float`
        Maximum wind speed in miles per hour
    wind_kph: :class:`float`
        Maximum wind speed in kilometer per hour
    wind_degree: :class:`int`
        Wind direction in degrees
    wind_dir: :class:`str`
        Wind direction as 16 point compass. e.g.: NSW
    pressure_mb: :class:`float`
        Pressure in millibars
    pressure_in: :class:`float`
        Pressure in inches
    precip_mm: :class:`float`
        Precipitation amount in millimeters
    precip_in: :class:`float`
        Precipitation amount in inches
    humidity: :class:`int`
        Humidity as percentage
    cloud: :class:`int`
        Cloud cover as percentage
    feelslike_c: :class:`float`
        Feels like temperature as celcius
    feelslike_f: :class:`float`
        Feels like temperature as fahrenheit
    windchill_c: :class:`float`
        Windchill temperature in celcius
    windchill_f: :class:`float`
        Windchill temperature in fahrenheit
    headindex_c: :class:`float`
        Heat index in celcius
    headindex_f: :class:`float`
        Heat index in fahrenheit
    dewpoint_c: :class:`float`
        Dew point in celcius
    dewpoint_f: :class:`float`
        Dew point in fahrenheit
    will_it_rain: :class:`bool`
        Will it will rain or not
    will_it_snow: :class:`bool`
        Will it will snow or not
    is_day: :class:`bool`
        Whether to show day condition icon or night icon
    vis_km: :class:`float`
        Visibility in kilometer
    vis_miles: :class:`float`
        Visibility in miles
    chance_of_rain: :class:`int`
        Chance of rain as percentage
    chance_of_snow: :class:`int`
        Chance of snow as percentage
    gust_mph: :class:`float`
        Wind gust in miles per hour
    gust_kph: :class:`float`
        Wind gust in kilometer per hour
    uv: Optional[:class:`float`]
        UV Index. Can be ``None`` when this class is returned from Future API
    aqi: Optional[:class:`AirQualityData`]
        Air Quality data. See :class:`AirQualityData` for more info.
    """
    def __init__(
        self,
        raw: Dict[str, Any],
        status: int,
        code: Optional[int]
    ) -> None:
        super().__init__(raw, status, code)

        self.time_epoch: int = raw["time_epoch"]
        self.time: str = raw["time"]
        self.temp_c: float = raw["temp_c"]
        self.temp_f: float = raw["temp_f"]
        self.condition_text: str = raw["condition"]["text"]
        self.condition_icon: str = raw["condition"]["icon"]
        self.condition_code: int = raw["condition"]["code"]
        self.wind_mph: float = raw["wind_mph"]
        self.wind_kph: float = raw["wind_kph"]
        self.wind_degree: int = raw["wind_degree"]
        self.wind_dir: str = raw["wind_dir"]
        self.pressure_mb: float = raw["pressure_mb"]
        self.pressure_in: float = raw["pressure_in"]
        self.precip_mm: float = raw["precip_mm"]
        self.precip_in: float = raw["precip_in"]
        self.humidity: int = raw["humidity"]
        self.cloud: int = raw["cloud"]
        self.feelslike_c: float = raw["feelslike_c"]
        self.feelslike_f: float = raw["feelslike_f"]
        self.windchill_c: float = raw["windchill_c"]
        self.windchill_f: float = raw["windchill_f"]
        self.heatindex_c: float = raw["heatindex_c"]
        self.heatindex_f: float = raw["heatindex_f"]
        self.dewpoint_c: float = raw["dewpoint_c"]
        self.dewpoint_f: float = raw["dewpoint_f"]
        self.will_it_rain: bool = bool(raw["will_it_rain"])
        self.will_it_snow: bool = bool(raw["will_it_snow"])
        self.is_day: bool = bool(raw["is_day"])
        self.vis_km: float = raw["vis_km"]
        self.vis_miles: float = raw["vis_miles"]
        self.chance_of_rain: int = raw["chance_of_rain"]
        self.chance_of_snow: int = raw["chance_of_snow"]
        self.gust_mph: float = raw["gust_mph"]
        self.gust_kph: float = raw["gust_kph"]
        self.uv: Optional[float] = raw.get("uv")
        
        self.aqi: Optional[AirQualityData] = None
        if raw.get("air_quality"): 
            self.aqi: Optional[AirQualityData] = AirQualityData(raw["air_quality"], status, code)
    
class ForecastDay(APIResponse):
    """
    A forecast day, element of :class:`ForecastData`

    Attributes
    -------------
    raw: Dict[:class:`str`, Any]
        Raw response in a JSON-like format (converted to a python dictionary)
    status: :class:`int`
        HTTP status of the response. 200 is OK, and is the most common status.
    code: Optional[:class:`int`]
        Response code. In some cases this can be ``None``
    date: :class:`str`
        Forecast date
    date_epoch: :class:`int`
        Forecast date as unix time.
    maxtemp_c: :class:`float`
        Maximum temperature in celsius for the day.
    maxtemp_f: :class:`float`
        Maximum temperature in fahrenheit for the day
    mintemp_c: :class:`float`
        Minimum temperature in celsius for the day
    mintemp_f: :class:`float`
        Minimum temperature in fahrenheit for the day
    avgtemp_c: :class:`float`
        Average temperature in celsius for the day
    avgtemp_f: :class:`float`
        Average temperature in fahrenheit for the day
    maxwind_mph: :class:`float`
        Maximum wind speed in miles per hour
    maxwind_kph: :class:`float`
        Maximum wind speed in kilometer per hour
    totalprecip_mm: :class:`float`
        Total precipitation in milimeter
    totalprecip_in: :class:`float`
        Total precipitation in inches
    avgvis_km: :class:`float`
        Average visibility in kilometer
    avgvis_miles: :class:`float`
        Average visibility in miles
    avghumidity: :class:`int`
        Average humidity as percentage
    uv: Optional[:class:`float`]
        UV Index. Can be ``None`` when this class is returned from Future API
    condition_text: :class:`str`
        Weather condition text
    condition_icon: :class:`str`
        Weather condition icon
    condition_code: :class:`int`
        Weather condition code
    hour_data: List[:class:`ForecastHour`]
        A list of :class:`ForecastHour` objects representing hourly weather data.
    astro: :class:`AstronomicalData`
        Astronomical data object
    aqi: Optional[:class:`AirQualityData`]
        Air Quality data as :class:`AirQualityData` object. Can be ``None``
    """
    def __init__(
        self,
        raw: Dict[str, Any],
        status: int,
        code: Optional[int]
    ) -> None:
        super().__init__(raw, status, code)
        self.date = raw["date"]
        self.date_epoch = raw["date_epoch"]
        
        before_raw = raw.copy()
        raw = raw["day"]
        
        self.maxtemp_c: float = raw["maxtemp_c"]
        self.maxtemp_f: float = raw["maxtemp_f"]
        self.mintemp_c: float = raw["mintemp_c"]
        self.mintemp_f: float = raw["mintemp_f"]
        self.avgtemp_c: float = raw["avgtemp_c"]
        self.avgtemp_f: float = raw["avgtemp_f"]
        self.maxwind_mph: float = raw["maxwind_mph"]
        self.maxwind_kph: float = raw["maxwind_kph"]
        self.totalprecip_in: float = raw["totalprecip_in"]
        self.totalprecip_mm: float = raw["totalprecip_mm"]
        self.avgvis_km: float = raw["avgvis_km"]
        self.avgvis_miles: float = raw["avgvis_miles"]
        self.avghumidity: int = raw["avghumidity"]
        self.uv: Optional[float] = raw.get("uv")
        self.condition_text: str = raw["condition"]["text"]
        self.condition_icon: str = raw["condition"]["icon"]
        self.condition_code: int = raw["condition"]["code"]
        if raw.get("air_quality"): self.aqi: AirQualityData = AirQualityData(raw["air_quality"], status, code)
        else: self.aqi = None
        
        raw = before_raw
        self.hour_data = list([
            ForecastHour(elem, status, code) for elem in raw["hour"]
        ])
        self.astro: AstronomicalData = AstronomicalData(raw["astro"], status, code)

    
class ForecastData(APIResponse):
    """
    Forecast data returned from Forecast API
    
    Attributes
    -------------
    raw: Dict[:class:`str`, Any]
        Raw response in a JSON-like format (converted to a python dictionary)
    status: :class:`int`
        HTTP status of the response. 200 is OK, and is the most common status.
    code: Optional[:class:`int`]
        Response code. In some cases this can be ``None``
    location: :class:`LocationData`
        Location of the forecast data.
    forecast_days: List[:class:`ForecastDay`]
        A list of forecast days
    alerts: List[:class:`AlertData`]
        A list of alerts, this list can be empty. List is also empty, when the user disabled alerts in the request.
    """
    def __init__(
        self,
        raw: Dict[str, Any],
        status: int,
        code: Optional[int]
    ) -> None:
        super().__init__(raw, status, code)
        
        self.location: LocationData = LocationData(raw["location"], status, code)
        self.forecast_days: List[ForecastDay] = list([ForecastDay(elem, status, code) for elem in raw["forecast"]["forecastday"]])
        self.alerts: List[AlertData] = []
        alert_data = raw.get("alerts", {})
        
        for k,v in alert_data.items():
            if v is not []:
                self.alerts.extend(list([
                    AlertData(elem, status, code) for elem in v
                ]))
    
    def iter_hours(self) -> Iterator[ForecastHour]:
        """
        Iterate over all possible hours from this class.

        .. container:: operations

            .. describe:: for x in y

                Iterate over hour data from this instance.

        Returns
        -------------
        Iterator[:class:`ForecastHour`]
            A generator object you can iterate over made of :class:`ForecastHour`
        """

        for day in self.forecast_days:
            for hour in day.hour_data:
                yield hour


class FutureData(APIResponse):
    """
    Forecast data returned from Future API
    
    Attributes
    -------------
    raw: Dict[:class:`str`, Any]
        Raw response in a JSON-like format (converted to a python dictionary)
    status: :class:`int`
        HTTP status of the response. 200 is OK, and is the most common status.
    code: Optional[:class:`int`]
        Response code. In some cases this can be ``None``
    location: :class:`LocationData`
        Location of the forecast data.
    day: :class:`ForecastDay`
        Day data for the requested future date.
    """
    def __init__(
        self,
        raw: Dict[str, Any],
        status: int,
        code: Optional[int]
    ) -> None:
        super().__init__(raw, status, code)
        
        self.location: LocationData = LocationData(raw["location"], status, code)
        self.day: ForecastDay = ForecastDay(raw["forecast"]["forecastday"][0], status, code)

class AstronomicalData(APIResponse):
    """
    Represents astronomical data as a response from WeatherAPI
    
    Attributes
    -----------
    raw: Dict[:class:`str`, Any]
        Raw response in a JSON-like format (converted to a python dictionary)
    status: :class:`int`
        HTTP status of the response. 200 is OK, and is the most common status.
    code: Optional[:class:`int`]
        Response code. In some cases this can be ``None``
    location: Optional[:class:`LocationData`]
        Location of the requested data. Is not ``None`` only if this object is returned from ``astronomy.json`` endpoint.
    sunrise: :class:`str`
        Sunrise local time
    sunset: :class:`str`
        Sunset local time
    moonrise: :class:`str`
        Moonrise local time
    moonset: :class:`str`
        Moonset local time
    moon_phase: :class:`str`
        Moon phases. Value returned:
            * New Moon
            * Waxing Crescent
            * First Quarter
            * Waxing Gibbous
            * Full Moon
            * Waning Gibbous
            * Last Quarter
            * Waning Crescent
    moon_illumination: :class:`int`
        Moon illumination
    """
    def __init__(
        self,
        raw: Dict[str, Any],
        status: int,
        code: Optional[int],    
    ) -> None:
        super().__init__(raw, status, code)
        
        # there are two types of "raw" data we need to convert
        # 1. from astronomy.json 
        # 2. from other types of endpoints, e.g. forecast data
        # we set location to None to avoid errors
        self.location: Optional[LocationData] = None
                
        if raw.get("astronomy") is not None:
            # first type
            self.location = LocationData(raw["location"], status, code)
            raw = raw["astronomy"]["astro"]
        
        self.sunrise: str = raw["sunrise"]
        self.sunset: str = raw["sunset"]
        self.moonrise: str = raw["moonrise"]
        self.moonset: str = raw["moonset"]
        self.moon_phase: str = raw["moon_phase"]
        self.moon_illumination: int = int(raw["moon_illumination"])
        
class IPData(APIResponse):
    """
    Represents IP Data retuned from IP Lookup API
    
    Attributes
    ------------
    raw: Dict[:class:`str`, Any]
        Raw response in a JSON-like format (converted to a python dictionary)
    status: :class:`int`
        HTTP status of the response. 200 is OK, and is the most common status.
    code: Optional[:class:`int`]
        Response code. In some cases this can be ``None``
    ip: :class:`str`
        IP address
    type: Literal["ipv4", "ipv6"]
        Type of the IP address
    continent_code: :class:`str`
        Continent code
    continent_name: :class:`str`
        Continent name
    country_code: :class:`str`
        Country code
    country_name: :class:`str`
        Name of country
    is_eu: :class:`bool`
        Indicating if IP address is in Europe
    geoname_id: :class:`int`
        Geoname ID
    city: :class:`str`
        City name
    region: :class:`str`
        Region name
    latitude: :class:`float`
        Latitude in decimal degree
    longitude: :class:`float`
        Longitude in decimal degree
    tz_id: :class:`str`
        Time zone
    localtime_epoch: :class:`int`
        Local time epoch
    localtime_formatted: :class:`str`
        Formatted local time string (e.g. 2023-04-30 17:54)
    """
    def __init__(
        self,
        raw: Dict[str, Any],
        status: int,
        code: Optional[int]
    ) -> None:
        super().__init__(raw, status, code)
        
        self.ip: str = raw["ip"]
        self.type: Literal["ipv4", "ipv6"] = raw["type"]
        self.continent_code: str = raw["continent_code"]
        self.continent_name: str = raw["continent_name"]
        self.country_code: str = raw["country_code"]
        self.country_name: str = raw["country_name"]
        self.is_eu: bool = bool(raw["is_eu"])
        self.geoname_id: int = raw["geoname_id"]
        self.city: str = raw["city"]
        self.region: str = raw["region"]
        self.latitude: float = raw["lat"]
        self.longitude: float = raw["lon"]
        self.tz_id: str = raw["tz_id"]
        self.localtime_epoch: int = raw["localtime_epoch"]
        self.localtime_formatted: str = raw["localtime"]

class SportsEvent(APIResponse):
    """
    Represents a single sports event from the Sports API

    Attributes
    ---------------
    raw: Dict[:class:`str`, Any]
        Raw response in a JSON-like format (converted to a python dictionary)
    status: :class:`int`
        HTTP status of the response. 200 is OK, and is the most common status.
    code: Optional[:class:`int`]
        Response code. In some cases this can be ``None``
    stadium: :class:`str`
        Name of stadium
    country: :class:`str`
        Country
    region: :class:`str`
        Region
    tournament: :class:`str`
        Tournament name
    start_time: :class:`str`
        Start local date and time for event in yyyy-MM-dd HH:mm format
    match: :class:`str` 
        Match name
    event_type: :class:`SportsEventType`
        An enum representing the type of event. Can be "golf", football" or "cricket"
    """
    def __init__(
        self, 
        raw: Dict[str, Any], 
        status: int, 
        code: Optional[int],
        event_type: SportsEventType
    ) -> None:
        super().__init__(raw, status, code)
        self.event_type = event_type
        self.stadium = raw["stadium"]
        self.country = raw["country"]
        self.region = raw["region"]
        self.tournament = raw["tournament"]
        self.start_time = raw["start"]
        self.match = raw["match"]

class SportsData(APIResponse):
    """
    Represents sports data returned from the Sports API

    Attributes
    -------------
    raw: Dict[:class:`str`, Any]
        Raw response in a JSON-like format (converted to a python dictionary)
    status: :class:`int`
        HTTP status of the response. 200 is OK, and is the most common status.
    code: Optional[:class:`int`]
        Response code. In some cases this can be ``None``
    events: List[:class:`SportsEvent`]
        A list of sports 
    categorized: Dict[str, List[:class:`SportsEvent`]]
        Dictionary of events, that are same as in :py:attr:`~events`, but categorized. 
        Category is the key, and value is a list of :class:`SportsEvent`.
    """
    def __init__(
        self,
        raw: Dict[str, Any],
        status: int,
        code: Optional[int]
    ) -> None:
        super().__init__(raw, status, code)

        self.events = []
        self.categorized = {"football": [], "golf": [], "cricket": []}
        for fball in raw["football"]:
            obj = SportsEvent(fball, status, code, SportsEventType.football)
            self.events.append(obj)
            self.categorized["football"].append(obj)

        for golf in raw["golf"]:
            obj = SportsEvent(golf, status, code, SportsEventType.golf)
            self.events.append(obj)
            self.categorized["golf"].append(obj)
        
        for cricket in raw["cricket"]:
            obj = SportsEvent(cricket, status, code, SportsEventType.cricket)
            self.events.append(obj)
            self.categorized["cricket"].append(obj)


"""
MIT License

Copyright (c) 2023 Konrad (@konradsic)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from typing import (
    Dict,
    Any,
    overload
)
from .enums import Languages

gray = "#535353"
white = "#ffffff"

def parse_kwargs_to_urlargs(kwargs: Dict[str, Any]) -> str:
    """
    Parses keyword arguments (kwargs) to format fitting URLs
    Example: ``{"some_key": "some value", "yes": "no"}`` -> ``?some_key=some_value&yes=no``
    
    Parameters
    ----------
    kwargs: Dict[:class:`str`, Any]
        A dictionary of keyword arguments to be parsed.
        
    Returns
    -------
    args_string: :class:`str`
        A string - formatted keyword arguments
    """
    if not kwargs: return ""
    
    args_string = ""
    first = True
    
    for k,v in kwargs.items():
        args_string += f"{'?' if first else '&'}{k}={v}"
        first = False
    return args_string

@overload
def find_language(lang: str, asobj: bool=False) -> None:
    ...
    
@overload
def find_language(lang: str, asobj: bool=False) -> Languages:
    ...

def find_language(lang: str, asobj: bool=False) -> str:
    """
    Used to find language code from available languages.

    Paremeters
    ----------
    lang: :class:`str`
        A language you want to search for

    Returns
    -------
    Union[:class:`str`, :class:`Languages`, None]
        Language code, could be ``None``, could be a :class:`Languages` enum.
    """

    if isinstance(lang, Languages):
        try:
            if asobj: return lang
            return lang.value
        except: 
            return None

    lang = lang.lower()

    for language in Languages:
        if language.name.lower() == lang or language.value == lang:
            if asobj: return language
            return language.value
        
    return None